@@ -19,7 +19,7 @@
 static int		bmap_f(int argc, char **argv);
 static int		bmap_one_extent(xfs_bmbt_rec_t *ep,
 					xfs_fileoff_t *offp, xfs_fileoff_t eoff,
-					int *idxp, bmap_ext_t *bep);
+					xfs_extnum_t *idxp, bmap_ext_t *bep);
 static xfs_fsblock_t	select_child(xfs_fileoff_t off, xfs_bmbt_key_t *kp,
 				     xfs_bmbt_ptr_t *pp, int nrecs);
 
@@ -32,7 +32,7 @@ bmap(
 	xfs_fileoff_t		offset,
 	xfs_filblks_t		len,
 	int			whichfork,
-	int			*nexp,
+	xfs_extnum_t		*nexp,
 	bmap_ext_t		*bep)
 {
 	struct xfs_btree_block	*block;
@@ -44,10 +44,10 @@ bmap(
 	enum xfs_dinode_fmt	fmt;
 	int			fsize;
 	xfs_bmbt_key_t		*kp;
-	int			n;
+	xfs_extnum_t		n;
 	int			nex;
 	xfs_fsblock_t		nextbno;
-	int			nextents;
+	xfs_extnum_t		nextents;
 	xfs_bmbt_ptr_t		*pp;
 	xfs_bmdr_block_t	*rblock;
 	typnm_t			typ;
@@ -68,7 +68,7 @@ bmap(
 	ASSERT(fmt == XFS_DINODE_FMT_LOCAL || fmt == XFS_DINODE_FMT_EXTENTS ||
 		fmt == XFS_DINODE_FMT_BTREE);
 	if (fmt == XFS_DINODE_FMT_EXTENTS) {
-		nextents = XFS_DFORK_NEXTENTS(dip, whichfork);
+		nextents = xfs_dfork_nextents(dip, whichfork);
 		xp = (xfs_bmbt_rec_t *)XFS_DFORK_PTR(dip, whichfork);
 		for (ep = xp; ep < &xp[nextents] && n < nex; ep++) {
 			if (!bmap_one_extent(ep, &curoffset, eoffset, &n, bep))
@@ -78,8 +78,8 @@ bmap(
 		push_cur();
 		rblock = (xfs_bmdr_block_t *)XFS_DFORK_PTR(dip, whichfork);
 		fsize = XFS_DFORK_SIZE(dip, mp, whichfork);
-		pp = XFS_BMDR_PTR_ADDR(rblock, 1, libxfs_bmdr_maxrecs(fsize, 0));
-		kp = XFS_BMDR_KEY_ADDR(rblock, 1);
+		pp = xfs_bmdr_ptr_addr(rblock, 1, libxfs_bmdr_maxrecs(fsize, 0));
+		kp = xfs_bmdr_key_addr(rblock, 1);
 		bno = select_child(curoffset, kp, pp,
 					be16_to_cpu(rblock->bb_numrecs));
 		for (;;) {
@@ -88,9 +88,9 @@ bmap(
 			block = (struct xfs_btree_block *)iocur_top->data;
 			if (be16_to_cpu(block->bb_level) == 0)
 				break;
-			pp = XFS_BMBT_PTR_ADDR(mp, block, 1,
+			pp = xfs_bmbt_ptr_addr(mp, block, 1,
 				libxfs_bmbt_maxrecs(mp, mp->m_sb.sb_blocksize, 0));
-			kp = XFS_BMBT_KEY_ADDR(mp, block, 1);
+			kp = xfs_bmbt_key_addr(mp, block, 1);
 			bno = select_child(curoffset, kp, pp,
 					be16_to_cpu(block->bb_numrecs));
 		}
@@ -98,7 +98,7 @@ bmap(
 			nextbno = be64_to_cpu(block->bb_u.l.bb_rightsib);
 			nextents = be16_to_cpu(block->bb_numrecs);
 			xp = (xfs_bmbt_rec_t *)
-				XFS_BMBT_REC_ADDR(mp, block, 1);
+				xfs_bmbt_rec_addr(mp, block, 1);
 			for (ep = xp; ep < &xp[nextents] && n < nex; ep++) {
 				if (!bmap_one_extent(ep, &curoffset, eoffset,
 						&n, bep)) {
@@ -132,7 +132,7 @@ bmap_f(
 	struct xfs_dinode	*dip;
 	xfs_fileoff_t		eo;
 	xfs_filblks_t		len;
-	int			nex;
+	xfs_extnum_t		nex;
 	char			*p;
 	int			whichfork;
 
@@ -158,9 +158,9 @@ bmap_f(
 		push_cur();
 		set_cur_inode(iocur_top->ino);
 		dip = iocur_top->data;
-		if (be32_to_cpu(dip->di_nextents))
+		if (xfs_dfork_data_extents(dip))
 			dfork = 1;
-		if (be16_to_cpu(dip->di_anextents))
+		if (xfs_dfork_attr_extents(dip))
 			afork = 1;
 		pop_cur();
 	}
@@ -223,13 +223,13 @@ bmap_one_extent(
 	xfs_bmbt_rec_t		*ep,
 	xfs_fileoff_t		*offp,
 	xfs_fileoff_t		eoff,
-	int			*idxp,
+	xfs_extnum_t		*idxp,
 	bmap_ext_t		*bep)
 {
 	xfs_filblks_t		c;
 	xfs_fileoff_t		curoffset;
 	int			f;
-	int			idx;
+	xfs_extnum_t		idx;
 	xfs_fileoff_t		o;
 	xfs_fsblock_t		s;
 
@@ -276,10 +276,10 @@ convert_extent(
 void
 make_bbmap(
 	bbmap_t		*bbmap,
-	int		nex,
+	xfs_extnum_t	nex,
 	bmap_ext_t	*bmp)
 {
-	int		i;
+	xfs_extnum_t	i;
 
 	for (i = 0; i < nex; i++) {
 		bbmap->b[i].bm_bn = XFS_FSB_TO_DADDR(mp, bmp[i].startblock);
